﻿import mx.transitions.OnEnterFrameBeacon;
import com.robertpenner.geom.Vector;
import com.robertpenner.physics.Force;

/**
* The PhysicsParticle class controls a MovieClip, 
* turning it into a particle driven by physics forces. 
* Designed to collaborate with the Force class (and subclasses).
* 
* @author Robert Penner (AS 1.0)
* © 2002 Robert Penner
* http://www.robertpenner.com
* 
* @author Mark Walters (AS 2.0)
* © 2006 DigitalFlipbook
* http://www.digitalflipbook.com
* 
* @version	1.0
* 
* @since	ActionScript 2.0; Flash Player 6
*/
class com.robertpenner.physics.PhysicsParticle {
	
	/*
	* CLASS PROPERTIES
	*/
	private var $target:MovieClip;
	private var $pos:Vector;
	private var $vel:Vector;
	private var $accel:Vector;
	private var $friction:Vector;
	private var $forces:Object;
	private var $xProp:String;
	private var $yProp:String;
	
	
	/*
	* *********************************************************
	* CONSTRUCTOR
	* *********************************************************
	*
	*/
	
	/**
	* Creates an instance of the PhysicsParticle class.
	* 
	* @param	target		The MovieClip that represents the particle visually.
	* @param	x			The initial x position of the particle.
	* @param	y			The initial y position of the particle.
	* @param	xProp		The name of the property to link to the x component of the position; this defaults to "_x" if the parameter is omitted.
	* @param	yProp		The name of the property to link to the y component of the position; this defaults to "_y" if the parameter is omitted.
	* 
	* @usage
	* <pre><code>
	* var p:PhysicsParticle = new PhysicsParticle(this, 0, 0, "_x", "_yscale");
	* </code></pre>
	*/
	function PhysicsParticle (target:MovieClip, x:Number, y:Number, xProp:String, yProp:String) {
		this.$target = target;
		this.$pos = new Vector (x, y);
		this.$vel = new Vector (0, 0);
		this.$accel = new Vector (0, 0);
		this.$friction = new Vector (0, 0);
		this.setXYProps (xProp, yProp);
		this.$forces = {};
		OnEnterFrameBeacon.init();
		MovieClip.addListener(this);
	}
	
	
	/*
	* *********************************************************
	* PRIVATE METHODS
	* *********************************************************
	* 
	*/
	
	private function move ():Void {
		var forceX:Number = 0
		var forceY:Number = 0;
		var fs:Object = this.$forces;
		for (var i in fs) {
			if (!fs[i].live()) {
				delete fs[i];
			} else {
				forceX += fs[i].value.x;
				forceY += fs[i].value.y;
			}
		}	
		this.$accel.reset (forceX, forceY);
		var v:Vector = this.$vel;
		v.plus (this.$accel);
		v.x *= (1 - this.$friction.x);
		v.y *= (1 - this.$friction.y);
		this.$pos.plus (v);
		this.render();
	}
	
	private function render ():Void {
		this.$target[this.$xProp] = this.$pos.x;
		this.$target[this.$yProp] = this.$pos.y;
	}
	
	private function onEnterFrame ():Void {
		this.move();
	}
	
	
	/*
	* *********************************************************
	* PUBLIC METHODS
	* *********************************************************
	* 
	*/
	
	/**
	* Adds a force to a PhysicsParticle instance.
	* 
	* @param	id			A string that gives the force a unique identifier.
	* @param	force		An instance of the Force class that will be applied to the particle.
	* 
	* @usage
	* <pre><code>
	* var p:PhysicsParticle = new PhysicsParticle(this, 0, 0, "_x", "_yscale");
	* var gravityForce:Force = new Force(0, .5);
	* p.addForce("gravity", gravityForce);
	* </code></pre>
	*/
	public function addForce (id:String, force:Force):Void {
		this.$forces[id] = force;
		force.parent = this;
	}
	
	/**
	* Removes a force from a PhysicsParticle instance.
	* 
	* @param	id			The string identifier of the force to remove.
	* 
	* @usage
	* <pre><code>
	* var p:PhysicsParticle = new PhysicsParticle(this, 0, 0, "_x", "_yscale");
	* var gravityForce:Force = new Force(0, .5);
	* p.addForce("gravity", gravityForce);
	* p.removeForce("gravity");
	* </code></pre>
	*/
	public function removeForce (id:String):Void {
		delete this.$forces[id];
	}
	
	/**
	* Gets the position of a PhysicsParticle instance.
	* 
	* @usage
	* <pre><code>
	* trace (p.getPosition());
	* </code></pre>
	*/
	public function getPosition ():Vector {
		return this.position;
	}
	
	/**
	* Sets the position of a PhysicsParticle instance.
	* 
	* @param	x			The x position of the PhysicsParticle.
	* @param	y			The y position of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* var p:PhysicsParticle = new PhysicsParticle(this, 0, 0, "_x", "_yscale");
	* p.setPosition(10, 10);
	* </code></pre>
	*/
	public function setPosition (x:Number, y:Number):Void {
		this.$pos.reset (x, y);
	}
	
	/**
	* Gets the velocity of a PhysicsParticle instance.
	* 
	* @usage
	* <pre><code>
	* trace (p.getVelocity());
	* </code></pre>
	*/
	public function getVelocity ():Vector {
		return this.velocity;
	}
	
	/**
	* Sets the velocity of a PhysicsParticle instance.
	* 
	* @param	vx			The x velocity of the PhysicsParticle.
	* @param	vy			The y velocity of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* var p:PhysicsParticle = new PhysicsParticle(this, 0, 0, "_x", "_yscale");
	* p.setVelocity(2, 0);
	* </code></pre>
	*/
	public function setVelocity (vx:Number, vy:Number):Void {
		this.$vel.reset (vx, vy);
	}
	
	/**
	* Gets the friction of a PhysicsParticle instance.
	* 
	* @usage
	* <pre><code>
	* trace (p.getFriction());
	* </code></pre>
	*/
	public function getFriction ():Vector {
		return this.friction;
	}
	
	/**
	* Sets the friction of a PhysicsParticle instance.
	* 
	* @param	fx			The amount of friction to apply to the xProp of the PhysicsParticle.
	* @param	fy			The amount of friction to apply to the yProp of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* var p:PhysicsParticle = new PhysicsParticle(this, 0, 0, "_x", "_yscale");
	* p.setFriction(.03, .001);
	* </code></pre>
	*/
	public function setFriction (fx:Number, fy:Number):Void {
		if (fy == undefined) fy = fx;
		this.$friction.reset (fx, fy);
	}
	
	/**
	* Sets both xProp and yProp of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* p.setXYProps("_xscale", "_yscale");
	* </code></pre>
	*/
	public function setXYProps (xp:String, yp:String):Void {
		this.xProp = xp;
		this.yProp = yp;
	}
	
	
	/*
	* *********************************************************
	* GETTERS/SETTERS
	* *********************************************************
	* 
	*/
	
	/**
	* The position property of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* trace (p.position);
	* </code></pre>
	*/
	public function get position ():Vector {
		return this.$pos;
	}
	
	/**
	* The velocity property of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* trace (p.velocity);
	* </code></pre>
	*/
	public function get velocity ():Vector {
		return this.$vel;
	}	
	
	/**
	* The friction property of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* trace (p.position);
	* </code></pre>
	*/
	public function get friction ():Vector {
		return this.$friction;
	}
	
	/**
	* The xProp property of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* trace (p.xProp);
	* </code></pre>
	*/
	public function get xProp ():String {
		return this.$xProp;
	}
	public function set xProp (xp:String):Void {
		this.$xProp = (xp == undefined) ? "_x" : xp;
	}
	
	/**
	* The yProp property of the PhysicsParticle.
	* 
	* @usage
	* <pre><code>
	* trace (p.yProp);
	* </code></pre>
	*/
	public function get yProp ():String {
		return this.$yProp;
	}
	public function set yProp (yp:String):Void {
		this.$yProp = (yp == undefined) ? "_y" : yp;
	}
	
}
